/**
 * Copyright (c) 2016 Razeware LLC
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

import SpriteKit
import CoreMotion

class GameScene: SKScene {
  
  // Private GameScene Properties
  
  let motionManager = CMMotionManager()
  
  var contentCreated = false
  
  // 1
  var invaderMovementDirection: InvaderMovementDirection = .right
  // 2
  var timeOfLastMove: CFTimeInterval = 0.0
  // 3
  let timePerMove: CFTimeInterval = 1.0
  
  enum InvaderType {
    case a
    case b
    case c
    
    static var size: CGSize {
      return CGSize(width: 24, height: 16)
    }
    
    static var name: String {
      return "invader"
    }
  }
  
  enum InvaderMovementDirection {
    case right
    case left
    case downThenRight
    case downThenLeft
    case none
  }
  
  let kInvaderGridSpacing = CGSize(width: 12, height: 12)
  let kInvaderRowCount = 6
  let kInvaderColCount = 6
  
  let kShipSize = CGSize(width: 30, height: 16)
  let kShipName = "ship"
  
  let kScoreHudName = "scoreHud"
  let kHealthHudName = "healthHud"
  
  // Object Lifecycle Management
  
  // Scene Setup and Content Creation
  override func didMove(to view: SKView) {
    
    if (!self.contentCreated) {
      self.createContent()
      self.contentCreated = true
      motionManager.startAccelerometerUpdates()
    }
  }
  
  func createContent() {
    physicsBody = SKPhysicsBody(edgeLoopFrom: frame)
    
    setupInvaders()
    setupShip()
    setupHud()
    
    // black space color
    self.backgroundColor = SKColor.black
  }
  
  func makeInvader(ofType invaderType: InvaderType) -> SKNode {
    // 1
    var invaderColor: SKColor
    
    switch(invaderType) {
    case .a:
      invaderColor = SKColor.red
    case .b:
      invaderColor = SKColor.green
    case .c:
      invaderColor = SKColor.blue
    }
    
    // 2
    let invader = SKSpriteNode(color: invaderColor, size: InvaderType.size)
    invader.name = InvaderType.name
    
    return invader
  }
  
  func setupInvaders() {
    // 1
    let baseOrigin = CGPoint(x: size.width / 3, y: size.height / 2)
    
    for row in 0..<kInvaderRowCount {
      // 2
      var invaderType: InvaderType
      
      if row % 3 == 0 {
        invaderType = .a
      } else if row % 3 == 1 {
        invaderType = .b
      } else {
        invaderType = .c
      }
      
      // 3
      let invaderPositionY = CGFloat(row) * (InvaderType.size.height * 2) + baseOrigin.y
      
      var invaderPosition = CGPoint(x: baseOrigin.x, y: invaderPositionY)
      
      // 4
      for _ in 1..<kInvaderRowCount {
        // 5
        let invader = makeInvader(ofType: invaderType)
        invader.position = invaderPosition
        
        addChild(invader)
        
        invaderPosition = CGPoint(
          x: invaderPosition.x + InvaderType.size.width + kInvaderGridSpacing.width,
          y: invaderPositionY
        )
      }
    }
  }
  
  func setupShip() {
    // 1
    let ship = makeShip()
    
    // 2
    ship.position = CGPoint(x: size.width / 2.0, y: kShipSize.height / 2.0)
    addChild(ship)
  }
  
  func makeShip() -> SKNode {
    let ship = SKSpriteNode(color: SKColor.green, size: kShipSize)
    ship.name = kShipName
    
    // 1
    ship.physicsBody = SKPhysicsBody(rectangleOf: ship.frame.size)

    // 2
    ship.physicsBody!.isDynamic = true

    // 3
    ship.physicsBody!.affectedByGravity = false

    // 4
    ship.physicsBody!.mass = 0.02
    
    return ship
  }
  
  func setupHud() {
    // 1
    let scoreLabel = SKLabelNode(fontNamed: "Courier")
    scoreLabel.name = kScoreHudName
    scoreLabel.fontSize = 25
    
    // 2
    scoreLabel.fontColor = SKColor.green
    scoreLabel.text = String(format: "Score: %04u", 0)
    
    // 3
    scoreLabel.position = CGPoint(
      x: frame.size.width / 2,
      y: size.height - (40 + scoreLabel.frame.size.height/2)
    )
    addChild(scoreLabel)
    
    // 4
    let healthLabel = SKLabelNode(fontNamed: "Courier")
    healthLabel.name = kHealthHudName
    healthLabel.fontSize = 25
    
    // 5
    healthLabel.fontColor = SKColor.red
    healthLabel.text = String(format: "Health: %.1f%%", 100.0)
    
    // 6
    healthLabel.position = CGPoint(
      x: frame.size.width / 2,
      y: size.height - (80 + healthLabel.frame.size.height/2)
    )
    addChild(healthLabel)
  }
  
  // Scene Update
  
  func moveInvaders(forUpdate currentTime: CFTimeInterval) {
    // 1
    if (currentTime - timeOfLastMove < timePerMove) {
      return
    }
    
    determineInvaderMovementDirection()
    
    // 2
    enumerateChildNodes(withName: InvaderType.name) { node, stop in
      switch self.invaderMovementDirection {
      case .right:
        node.position = CGPoint(x: node.position.x + 10, y: node.position.y)
      case .left:
        node.position = CGPoint(x: node.position.x - 10, y: node.position.y)
      case .downThenLeft, .downThenRight:
        node.position = CGPoint(x: node.position.x, y: node.position.y - 10)
      case .none:
        break
      }
      
      // 3
      self.timeOfLastMove = currentTime
    }
  }
  
  func processUserMotion(forUpdate currentTime: CFTimeInterval) {
    // 1
    if let ship = childNode(withName: kShipName) as? SKSpriteNode {
      // 2
      if let data = motionManager.accelerometerData {
        // 3
        if fabs(data.acceleration.x) > 0.2 {
          // 4 How do you move the ship?
          print("Acceleration: \(data.acceleration.x)")
          ship.physicsBody!.applyForce(CGVector(dx: 40 * CGFloat(data.acceleration.x), dy: 0))
        }
      }
    }
  }
  
  override func update(_ currentTime: TimeInterval) {
    processUserMotion(forUpdate: currentTime)
    moveInvaders(forUpdate: currentTime)
  }
  
  // Scene Update Helpers
  
  // Invader Movement Helpers
  
  func determineInvaderMovementDirection() {
    // 1
    var proposedMovementDirection: InvaderMovementDirection = invaderMovementDirection
    
    // 2
    enumerateChildNodes(withName: InvaderType.name) { node, stop in
      
      switch self.invaderMovementDirection {
      case .right:
        //3
        if (node.frame.maxX >= node.scene!.size.width - 1.0) {
          proposedMovementDirection = .downThenLeft
          
          stop.pointee = true
        }
      case .left:
        //4
        if (node.frame.minX <= 1.0) {
          proposedMovementDirection = .downThenRight
          
          stop.pointee = true
        }
        
      case .downThenLeft:
        proposedMovementDirection = .left
        
        stop.pointee = true
        
      case .downThenRight:
        proposedMovementDirection = .right
        
        stop.pointee = true
        
      default:
        break
      }
      
    }
    
    //7
    if (proposedMovementDirection != invaderMovementDirection) {
      invaderMovementDirection = proposedMovementDirection
    }
  }
  
  // Bullet Helpers
  
  // User Tap Helpers
  
  // HUD Helpers
  
  // Physics Contact Helpers
  
  // Game End Helpers
  
}
